#undef _FILE_OFFSET_BITS
#define _FILE_OFFSET_BITS 64

#define _LARGEFILE_SOURCE

#define O_LARGEFILE 0

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <math.h>

#include "format.h"
#include "nuvread.h"


#define NELEMS(x)   (sizeof(x) / sizeof(x[0]))
#define BYTE(val,num) ((char)(((val) >> (8 * (num))) & 0xFF))
#define THRESH 0.0001

static int doubleEq( double a, double b );
static void double2fraction( double in, fraction *out );
static void nuv_closest_samplerate( nuvFile *nv );
static void nuv_closest_bitrate( nuvFile *nv );

int doubleEq( double a, double b )
{
    return( fabs(a - b) <= THRESH );
}

void double2fraction( double in, fraction *out )
{
    if( doubleEq(in, 29.97) ) {
        out->num = 30000;
        out->denom = 1001;
    } else if( doubleEq(in, 23.976) ) {
        out->num = 24000;
        out->denom = 1001;
    } else {
        out->denom = 1;
        while( in - floor(in) >= 0.1 ) {
            out->denom *= 10;
            in *= 10.0;
        }
        out->num = (int)floor(in);
    }
}

nuvFile *nuv_init( char *filename, int verbose )
{
    nuvFile        *nv;
    rtfileheader    rtFileHeader;
    extendeddata   *extradata;
    nuvFrame       *frame;
    int             bytes;
    char            string[17], text;
    int             i;

    nv = (nuvFile *)malloc(sizeof(nuvFile));
    if( !nv ) {
        fprintf( stderr, "No memory to create nuvFile!\n" );
        return( NULL );
    }

    nv->fd = open( filename, O_RDONLY|O_LARGEFILE );
    if( nv->fd < 0 ) {
        perror( "Unable to open file" );
        nuv_free_file(nv);
        return( NULL );
    }

    bytes = read(nv->fd, &rtFileHeader, FILEHEADERSIZE);
    if( bytes != FILEHEADERSIZE ) {
        fprintf( stderr, "Not enough data for a NUV header (%d bytes read)\n",
                 bytes );
        nuv_free_file(nv);
        return( NULL );
    }

    printf( "Filename: %s\n"
            "  Frame Size: %dx%d  FPS: %6.3f  Keyframe Distance: %d\n",
            filename, rtFileHeader.width, rtFileHeader.height,
            rtFileHeader.fps, rtFileHeader.keyframedist );

    nv->width      = rtFileHeader.width;
    nv->height     = rtFileHeader.height;
    nv->gop_length = rtFileHeader.keyframedist;
    double2fraction( rtFileHeader.fps, &nv->fps );

    frame = nuv_read_frame( nv, "D" );
    if( !frame ) {
        fprintf( stderr, "Missing initial D frame\n" );
        nuv_free_file(nv);
        return( NULL );
    }

    if( verbose ) {
        nuv_dump_frame(nv, frame);
    }

    frame = nuv_read_frame( nv, "X" );
    if( !frame || frame->length != EXTENDEDSIZE ) {
        fprintf( stderr, "Corrupt file: bad extended frame\n" );
        nuv_free_file(nv);
        return( NULL );
    }

    nuv_dump_frame(nv, frame);

    extradata = (extendeddata *)frame->data;

    nv->video_4cc               = extradata->video_fourcc;
    nv->audio_4cc               = extradata->audio_fourcc;
    nv->audio_sample_rate       = extradata->audio_sample_rate; 
    nv->audio_sample_rate_file  = extradata->audio_sample_rate; 
    nv->audio_bits_per_sample   = extradata->audio_bits_per_sample;
    nv->audio_channels          = extradata->audio_channels;
    nv->audio_compression_ratio = extradata->audio_compression_ratio;
    nv->audio_quality           = extradata->audio_quality;
    nv->video_bitrate           = extradata->lavc_bitrate;
    nv->seektable_offset        = extradata->seektable_offset;
    nv->kfa_offset              = extradata->keyframeadjust_offset;

    nuv_closest_samplerate( nv );
    nuv_closest_bitrate( nv );

    nuv_free_frame( frame );

    nv->frameCount = 0;
    nv->lastVideo  = -1;
    nv->lastAudio  = -1;
    nv->verbose    = verbose;

    return( nv );
}

nuvFrame *nuv_read_frame( nuvFile *nv, char *searchList )
{
    nuvFrame       *frame;
    char            frameType[2];

    while( 1 ) {
        frame = (nuvFrame *)malloc(sizeof(nuvFrame));
        if( !frame ) {
            fprintf( stderr, "Out of memory reading frame header!\n" );
            return( NULL );
        }

        if( read(nv->fd, &frame->header, FRAMEHEADERSIZE) != FRAMEHEADERSIZE ) {
            /* EOF */
            return( NULL );
        }

        frame->length = frame->header.packetlength;
        frame->data = NULL;

        if( frame->header.frametype != 'R' && frame->length ) {
            frame->data = (char *)malloc(frame->length);
            if( !frame->data ) {
                fprintf( stderr, "Out of memory reading frame body!\n" );
                nuv_free_frame( frame );
                return( NULL );
            }

            if( read( nv->fd, frame->data, frame->length ) != frame->length ) {
                /* premature EOF */
                nuv_free_frame( frame );
                return( NULL );
            }
        }

        frameType[0] = frame->header.frametype;
        frameType[1] = '\0';

        if( searchList && strstr(searchList, frameType) ) {
            nv->frameCount++;
            return( frame );
        }

        /* Ignore this puppy */
        if( nv->verbose ) {
            printf( "Skipping %s frame\n", frameType );
        }
        nuv_free_frame( frame );
    }
}

void nuv_free_frame( nuvFrame *frame )
{
    if( !frame ) {
        return;
    }

    if( frame->data ) {
        free( frame->data );
    }
    free( frame );
}

nuvFrame *nuv_duplicate_frame( nuvFrame *frame )
{
    nuvFrame       *newFrame;

    if( !frame ) {
        return( NULL );
    }

    newFrame = (nuvFrame *)malloc(sizeof(nuvFrame));
    if( !newFrame ) {
        return( NULL );
    }

    memcpy( newFrame, frame, sizeof(nuvFrame) );
    if( newFrame->length ) {
        newFrame->data = (char *)malloc(newFrame->length);
    } else {
        newFrame->data = NULL;
    }

    return( newFrame );
}

void nuv_free_file( nuvFile *nv )
{
    if( !nv ) {
        return;
    }

    close( nv->fd );
    free( nv );
}


void nuv_dump_frame( nuvFile *nv, nuvFrame *frame ) 
{
    extendeddata   *extradata;

    if( !frame || !nv ) {
        return;
    }

    switch( frame->header.frametype ) {
    case 'A':
        if( nv->lastAudio == -1 ) {
            printf( "Audio: Type: %c              Filter: %3d  "
                    "Time: %d.%03d  Len: %d\n",
                    frame->header.comptype, frame->header.filters, 
                    frame->header.timecode / 1000,
                    frame->header.timecode % 1000, frame->length );
        } else {
            printf( "Audio: Type: %c              Filter: %3d  "
                    "Time: %d.%03d  (%c%d.%03d)  Len: %d\n",
                    frame->header.comptype, frame->header.filters, 
                    frame->header.timecode / 1000,
                    frame->header.timecode % 1000, 
                    (frame->header.timecode > nv->lastAudio ? '+' :
                     (frame->header.timecode < nv->lastAudio ? '-' : ' ')),
                    abs(frame->header.timecode - nv->lastAudio) / 1000,
                    abs(frame->header.timecode - nv->lastAudio) % 1000, 
                    frame->length );
        }
        nv->lastAudio = frame->header.timecode;
        break;
    case 'V':
        if( nv->lastVideo == -1 ) {
            printf( "Video: Type: %c  FinGOP: %2d  Filter: %3d  "
                    "Time: %d.%03d  Len: %d\n",
                    frame->header.comptype, frame->header.keyframe,
                    frame->header.filters, 
                    frame->header.timecode / 1000,
                    frame->header.timecode % 1000, frame->length );
        } else {
            printf( "Video: Type: %c  FinGOP: %2d  Filter: %3d  "
                    "Time: %d.%03d  (%c%d.%03d)  Len: %d\n",
                    frame->header.comptype, frame->header.keyframe,
                    frame->header.filters, 
                    frame->header.timecode / 1000,
                    frame->header.timecode % 1000,
                    (frame->header.timecode > nv->lastVideo ? '+' :
                     (frame->header.timecode < nv->lastVideo ? '-' : ' ')),
                    abs(frame->header.timecode - nv->lastVideo) / 1000,
                    abs(frame->header.timecode - nv->lastVideo) % 1000,
                    frame->length );
        }
        nv->lastVideo = frame->header.timecode;
        break;
    case 'S':
        printf( "Sync:  Type: %c  Time: %d  Length: %d\n",
                frame->header.comptype, frame->header.timecode,
                frame->length );
        break;
    case 'T':
        printf( "Text:  Type: %c  Time: %d.%03d  Length: %d\n",
                frame->header.comptype, 
                frame->header.timecode / 1000, frame->header.timecode % 1000,
                frame->length );

        if( frame->length ) {
            nuv_hex_dump( frame );
        }
        break;
    case 'R':
        printf( "\nSeekpoint\n" );
        break;
    case 'D':
        printf( "Data:  Type: %c  Length: %d\n", 
                frame->header.comptype, frame->length );

        if( frame->length ) {
            nuv_hex_dump( frame );
        }
        break;
    case 'X':
        if( frame->length != EXTENDEDSIZE ) {
            fprintf( stderr, "Corrupt file: bad extended frame\n" );
            return;
        }

        extradata = (extendeddata *)frame->data;

        printf( "\nExtend: Version: %d  Video 4CC: %c%c%c%c "
                "Audio 4CC: %c%c%c%c\n", extradata->version,
                BYTE(extradata->video_fourcc, 0),
                BYTE(extradata->video_fourcc, 1),
                BYTE(extradata->video_fourcc, 2),
                BYTE(extradata->video_fourcc, 3),
                BYTE(extradata->audio_fourcc, 0),
                BYTE(extradata->audio_fourcc, 1),
                BYTE(extradata->audio_fourcc, 2),
                BYTE(extradata->audio_fourcc, 3) );
        printf( "   Audio: Rate: %d  Bits: %d  Channels: %d  "
                "Compr. Ratio: %d  Quality: %d\n", 
                extradata->audio_sample_rate, 
                extradata->audio_bits_per_sample,
                extradata->audio_channels,
                extradata->audio_compression_ratio,
                extradata->audio_quality );
        printf( "   RTJPEG: Quality: %d  Luma Filter: %d  "
                "Chroma Filter: %d\n", extradata->rtjpeg_quality,
                extradata->rtjpeg_luma_filter,
                extradata->rtjpeg_chroma_filter );
        printf( "   LAVC: Bitrate: %d  Qmin: %d  Qmax: %d  "
                "Max Qdiff: %d\n",
                extradata->lavc_bitrate, extradata->lavc_qmin,
                extradata->lavc_qmax, extradata->lavc_maxqdiff );
        printf( "   Offsets: seektable: %lld  KFA: %lld\n\n",
                extradata->seektable_offset,
                extradata->keyframeadjust_offset );
        break;
    case 'Q':
        printf( "Seek: Length: %d\n", frame->length );
        break;
    case 'K':
        printf( "KFA: Length: %d\n", frame->length );
        break;
    default:
        printf( "Unknown Frame Type: '%s'  Length: %d\n", 
                frame->header.frametype, frame->length );
        break;
    }
}

void nuv_hex_dump( nuvFrame *frame )
{
    int         i;
    char        string[17];

    for( i = 0; i < frame->length; i++ ) {
        if( i % 16 == 0 ) {
            printf( "  %04X: ", i );
            strcpy( string, "" );
        }

        printf( "%02X ", frame->data[i] & 0xFF );

        if( isprint( (int)frame->data[i] ) ) {
            string[i%16] = frame->data[i];
        } else {
            string[i%16] = '.';
        }
        string[(i%16) + 1] = '\0';

        if( i % 16 == 15 ) {
            printf( " | \"%s\"\n", string );
        }
    }

    if( i % 16 != 0 ) {
        printf( "%*s | \"%s\"\n", ((16 - (i % 16)) * 3), "", string );
    }
}

void nuv_closest_bitrate( nuvFile *nv )
{
    int         bitrate;
    int         i;
    const int   bitrate_table[] = {8000, 16000, 24000, 32000, 40000, 48000, 
                                   56000, 64000, 80000, 96000, 112000, 128000,
                                   160000, 192000, 224000, 256000, 320000};
    const int   ratecount = NELEMS(bitrate_table);

    bitrate = (nv->audio_sample_rate * nv->audio_bits_per_sample *
               nv->audio_channels) / nv->audio_compression_ratio;

    nv->audio_bitrate_calc = bitrate;

    if( bitrate <= bitrate_table[0] ) {
        bitrate = bitrate_table[0];
    } else if( bitrate >= bitrate_table[ratecount-1] ) {
        bitrate = bitrate_table[ratecount - 1];
    } else {
        for( i = 1; i < ratecount; i++ ) {
            if( bitrate <= bitrate_table[i] && 
                bitrate >= bitrate_table[i-1] ) {
                /* it's between the two, pick the closer */
                if( bitrate - bitrate_table[i-1] < 
                    bitrate_table[i] - bitrate ) {
                    bitrate = bitrate_table[i-1];
                } else {
                    bitrate = bitrate_table[i];
                }
                break;
            }
        }
    }

    nv->audio_bitrate_closest = bitrate;
}

void nuv_closest_samplerate( nuvFile *nv )
{
    int         sample;
    int         i;
    const int   samplerate_table[] = {32000, 44100, 48000};
    const int   ratecount = NELEMS(samplerate_table);

    sample = nv->audio_sample_rate_file;

    if( sample <= samplerate_table[0] ) {
        sample = samplerate_table[0];
    } else if( sample >= samplerate_table[ratecount - 1] ) {
        sample = samplerate_table[ratecount - 1];
    } else {
        for( i = 1; i < ratecount; i++ ) {
            if( sample <= samplerate_table[i] && 
                sample >= samplerate_table[i-1] ) {
                /* it's between the two, pick the closer */
                if( sample - samplerate_table[i-1] < 
                    samplerate_table[i] - sample ) {
                    sample = samplerate_table[i-1];
                } else {
                    sample = samplerate_table[i];
                }
                break;
            }
        }
    }

    nv->audio_sample_rate = sample;
}


/*
 * vim:ts=4:sw=4:ai:et:si:sts=4
 */
