
/*HEADER---------------------------------------------------
* $Id: main.c,v 1.4 2004/09/15 05:59:46 gjhurlbu Exp $
*
* Copyright 2004 Gavin Hurlbut
* All rights reserved
*
* Comments :
*
* Revision History:
* $Log: main.c,v $
* Revision 1.4  2004/09/15 05:59:46  gjhurlbu
* - added graceful aborting
* - added a GOP length queue in the reencoder
* - added status update in the writer
*
* Revision 1.3  2004/09/13 06:50:04  gjhurlbu
* - trimmed down the initialization of ffmpeg stuff to use only codecs, etc
*   we are using
*
* Revision 1.2  2004/09/13 06:24:56  gjhurlbu
* - runs again, now threaded
*
* Revision 1.1  2004/09/13 05:29:19  gjhurlbu
* - converted to threaded
*
*
*
*----------------------------------------------------------
*/

/* INCLUDE FILES */
#include "environment.h"
#include "queue.h"
#include "logging.h"
#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <errno.h>

#include "avformat.h"
#include "avcodec.h"

/* INTERNAL CONSTANT DEFINITIONS */

/* INTERNAL TYPE DEFINITIONS */

/* INTERNAL MACRO DEFINITIONS */
#define thread_create(a,b,c,d)  pthread_create((a),(b),(c),(d))

static pthread_t mainThreadId;
static pthread_t readerThreadId;
static pthread_t reencodeThreadId;
static pthread_t writerThreadId;

static char *inputFileName;
static char *outputFileName;

QueueObject_t *DecodeQueue;
QueueObject_t *WritingQueue;

bool GlobalAbort = FALSE;

void *ReaderThread( void *arg );
void *ReencodeThread( void *arg );
void *WriterThread( void *arg );

/* INTERNAL FUNCTION PROTOTYPES */
void LogBanner( void );
void MainParseArgs( int argc, char **argv );
void MainDisplayUsage( char *program, char *errorMsg );
void MainDelayReset( void );
void MainExit( int sigNum );

/* CVS generated ID string */
static char ident[] _UNUSED_= 
    "$Id: main.c,v 1.4 2004/09/15 05:59:46 gjhurlbu Exp $";

/*FUNCTION-------------------------------------------------
*
* Function Name :
* Returned Value :
* Parameters :
* Comments :
*
*----------------------------------------------------------
*/

int main ( int argc, char **argv )
{
    void *threadResult;
    int priority;
    int pid;

    GlobalAbort = FALSE;

    /* Parse the command line options */
    MainParseArgs( argc, argv );

    mainThreadId = pthread_self();

    /* Setup the exit handler */
    atexit( MainDelayReset );

    /* Setup the signal handler */
    signal( SIGINT, MainExit );

    /* Ignore SIGPIPE signals!  This allows TCP sockets to close while we're
     * writing to them.
     */
    signal( SIGPIPE, SIG_IGN );

    /* Print the startup log messages */
    LogBanner();

    /* Setup the Decode Queue */
    DecodeQueue = QueueCreate( 64 );

    /* Setup the Writing Queue */
    WritingQueue = QueueCreate( 64 );

    
    /* Initialize the AVI stuff */
    avcodec_init();
    register_avcodec(&mp3lame_encoder);
    register_avcodec(&mpeg4_encoder);
    register_avcodec(&mjpeg_encoder);
    avienc_init();
    register_protocol(&file_protocol);

    /* Start the Reading thread */
    thread_create( &readerThreadId, NULL, ReaderThread, (void *)inputFileName );

    /* Start the Reencoding thread */
    thread_create( &reencodeThreadId, NULL, ReencodeThread, NULL );

    /* Start the Writing thread */
    thread_create( &writerThreadId, NULL, WriterThread, 
                   (void *)outputFileName );

    /* Wait for Writing thread to finish */
    pthread_join( writerThreadId, &threadResult );

    /* Wait for Reencoding thread to finish */
    pthread_join( reencodeThreadId, &threadResult );

    /* Wait for Reading thread to finish */
    pthread_join( readerThreadId, &threadResult );

    return(0);
}


void LogBanner( void )
{
    LogPrint( LOG_CRIT, "nuv2avi  (c) 2004 Gavin Hurlbut\n" );
    LogPrint( LOG_CRIT, "Starting, $Revision: 1.4 $\n" );
}


void MainParseArgs( int argc, char **argv )
{
    extern char *optarg;
    extern int optind, opterr, optopt;
    int opt;

    inputFileName   = NULL;
    outputFileName  = NULL;

    while( (opt = getopt( argc, argv, "hi:o:" )) != -1 )
    {
        switch( opt )
        {
            case 'h':
                MainDisplayUsage( argv[0], NULL );
                exit( 0 );
                break;
            case 'i':
                if( inputFileName != NULL )
                {
                    free( inputFileName );
                }
                inputFileName = strdup(optarg);
                break;
            case 'o':
                if( outputFileName != NULL )
                {
                    free( outputFileName );
                }
                outputFileName = strdup(optarg);
                break;
            case '?':
            case ':':
            default:
                MainDisplayUsage( argv[0], "Unknown option" );
                exit( 1 );
                break;
        }
    }

    if( inputFileName == NULL || outputFileName == NULL )
    {
        MainDisplayUsage( argv[0], "You must specify both input and output "
                                   "files." );
        exit( 1 );
    }
}

void MainDisplayUsage( char *program, char *errorMsg )
{
    char *nullString = "<program name>";

    fprintf( stderr, "\nnuv2avi  (c) 2004 Gavin Hurlbut\n" );

    if( errorMsg != NULL )
    {
        fprintf( stderr, "\n%s\n\n", errorMsg );
    }

    if( program == NULL )
    {
        program = nullString;
    }

    fprintf( stderr, "\nUsage:\n\t%s -i inputFile -o outputFile\n",
             program );
}

void MainDelayReset( void )
{
    LogPrint( LOG_CRIT, "Shutting down\n" );

    /* Signal to all that we are aborting */
    GlobalAbort = TRUE;

    /* Delay so the threads can see the change */
    usleep( 300000 );

    /* Delay to allow all the other tasks to finish (esp. logging!) */
    usleep( 750000 );

#ifdef PROFILING
    _exit( 0 );
#endif
}

void MainExit( int sigNum )
{
    extern const char *const sys_siglist[];

    if( pthread_equal( pthread_self(), mainThreadId ) )
    {
        LogPrint( LOG_CRIT, "\nReceived signal: %s\n", sys_siglist[sigNum] );
        exit( 0 );
    }
}



/*
 * vim:ts=4:sw=4:ai:et:si:sts=4
 */

